package fr.jelora.jraspitools.display;

import fr.jelora.jraspitools.RasPiPin;
import fr.jelora.jraspitools.gpio.GPIODirection;
import fr.jelora.jraspitools.gpio.GPIOException;
import fr.jelora.jraspitools.gpio.GPIOTools;

public class GenericTextDisplay {
	
	private int cols;
	private int lines;
	
	private RasPiPin pinRegisterSelect;
	private RasPiPin pinReadWrite;
	private RasPiPin pinEnable;
	private RasPiPin pinData0;
	private RasPiPin pinData1;
	private RasPiPin pinData2;
	private RasPiPin pinData3;
	private RasPiPin pinData4;
	private RasPiPin pinData5;
	private RasPiPin pinData6;
	private RasPiPin pinData7;
	private DisplayDataBusMode displayDataBusMode;
	
	private GPIOTools gpioTools;
	
	private boolean displayOnOff;
	private boolean cursorPositionOnOff;
	private boolean cursorBlinkingOnOff;
	private boolean ignoreIfBusy;
	
	public GenericTextDisplay(int lines, int cols, RasPiPin pinRegisterSelect, RasPiPin pinReadWrite, RasPiPin pinEnable, 
			RasPiPin pinData0, RasPiPin pinData1, RasPiPin pinData2, RasPiPin pinData3,
			RasPiPin pinData4, RasPiPin pinData5, RasPiPin pinData6, RasPiPin pinData7,
			boolean ignoreIfBusy) throws DisplayException{
		if ( displayDataBusMode == null ){
			displayDataBusMode = DisplayDataBusMode.Mode8bits;
		}
		this.init(lines, cols, pinRegisterSelect, pinReadWrite, pinEnable, pinData0, pinData1, pinData2, pinData3, pinData4, pinData5, pinData6, pinData7, ignoreIfBusy);
	}
	
	public GenericTextDisplay(int lines, int cols, RasPiPin pinRegisterSelect, RasPiPin pinReadWrite, RasPiPin pinEnable, 
			RasPiPin pinData0, RasPiPin pinData1, RasPiPin pinData2, RasPiPin pinData3,
			RasPiPin pinData4, RasPiPin pinData5, RasPiPin pinData6, RasPiPin pinData7) throws DisplayException{
		if ( displayDataBusMode == null ){
			displayDataBusMode = DisplayDataBusMode.Mode8bits;
		}
		this.init(lines, cols, pinRegisterSelect, pinReadWrite, pinEnable, pinData0, pinData1, pinData2, pinData3, pinData4, pinData5, pinData6, pinData7, false);
	}
	
	public GenericTextDisplay(int lines, int cols, RasPiPin pinRegisterSelect, RasPiPin pinReadWrite, RasPiPin pinEnable, 
			RasPiPin pinData4, RasPiPin pinData5, RasPiPin pinData6, RasPiPin pinData7,
			boolean ignoreIfBusy) throws DisplayException{
		if ( displayDataBusMode == null ){
			displayDataBusMode = DisplayDataBusMode.Mode4bits;
		}
		this.init(lines, cols, pinRegisterSelect, pinReadWrite, pinEnable, null, null, null, null, pinData4, pinData5, pinData6, pinData7, ignoreIfBusy);
	}
	
	public GenericTextDisplay(int lines, int cols, RasPiPin pinRegisterSelect, RasPiPin pinReadWrite, RasPiPin pinEnable, 
			RasPiPin pinData4, RasPiPin pinData5, RasPiPin pinData6, RasPiPin pinData7) throws DisplayException{
		if ( displayDataBusMode == null ){
			displayDataBusMode = DisplayDataBusMode.Mode4bits;
		}
		this.init(lines, cols, pinRegisterSelect, pinReadWrite, pinEnable, null, null, null, null, pinData4, pinData5, pinData6, pinData7, false);
	}
	
	private void init(int lines, int cols, RasPiPin pinRegisterSelect, RasPiPin pinReadWrite, RasPiPin pinEnable, 
			RasPiPin pinData0, RasPiPin pinData1, RasPiPin pinData2, RasPiPin pinData3,
			RasPiPin pinData4, RasPiPin pinData5, RasPiPin pinData6, RasPiPin pinData7,
			boolean ignoreIfBusy) throws DisplayException{
		gpioTools = new GPIOTools();
		
		this.cols = cols;
		if ( this.cols < 1 || this.cols > 64 ){
			throw new DisplayException("wrong number of cols");
		}
		
		this.lines = lines;
		if ( this.lines != 1 && this.lines != 2 && this.lines != 4 ){
			throw new DisplayException("support displays with only 1/2/4 lines");
		}
		
		this.ignoreIfBusy = ignoreIfBusy;
		
		try{
			this.pinRegisterSelect = pinRegisterSelect;
			if ( this.pinRegisterSelect != null ){
				gpioTools.init(this.pinRegisterSelect, GPIODirection.out, false, false);
			}
			else{
				throw new DisplayException("pinRegisterSelect not selected");
			}
			
			this.pinReadWrite = pinReadWrite;
			if ( this.pinReadWrite != null ){
				gpioTools.init(this.pinReadWrite, GPIODirection.out, false, false);
			}
			
			this.pinEnable = pinEnable;
			if ( this.pinEnable != null ){
				gpioTools.init(this.pinEnable, GPIODirection.out, false, false);
			}
			else{
				throw new DisplayException("pinEnable not selected");
			}
			
			this.pinData0 = pinData0;
			if ( this.pinData0 != null ){
				gpioTools.init(this.pinData0, GPIODirection.out, false, false);
			}
			
			this.pinData1 = pinData1;
			if ( this.pinData1 != null ){
				gpioTools.init(this.pinData1, GPIODirection.out, false, false);
			}
			
			this.pinData2 = pinData2;
			if ( this.pinData2 != null ){
				gpioTools.init(this.pinData2, GPIODirection.out, false, false);
			}
			
			this.pinData3 = pinData3;
			if ( this.pinData3 != null ){
				gpioTools.init(this.pinData3, GPIODirection.out, false, false);
			}
			
			this.pinData4 = pinData4;
			if ( this.pinData4 != null ){
				gpioTools.init(this.pinData4, GPIODirection.out, false, false);
			}
			else{
				throw new DisplayException("pinData4 not selected");
			}
			
			this.pinData5 = pinData5;
			if ( this.pinData5 != null ){
				gpioTools.init(this.pinData5, GPIODirection.out, false, false);
			}
			else{
				throw new DisplayException("pinData5 not selected");
			}
			
			this.pinData6 = pinData6;
			if ( this.pinData6 != null ){
				gpioTools.init(this.pinData6, GPIODirection.out, false, false);
			}
			else{
				throw new DisplayException("pinData6 not selected");
			}
			
			this.pinData7 = pinData7;
			if ( this.pinData7 != null ){
				gpioTools.init(this.pinData7, GPIODirection.out, false, false);
			}
			else{
				throw new DisplayException("pinData7 not selected");
			}
		}
		catch (GPIOException e) {
			throw new DisplayException("error during init GPIOs", e);
		}
		
		if ( displayDataBusMode == DisplayDataBusMode.Mode8bits && (pinData0 == null || pinData1 == null || pinData2 == null || pinData3 == null ||
				pinData4 == null || pinData5 == null || pinData6 == null || pinData7 == null) ){
			throw new DisplayException("databus mode 8bits: not all 8 pins needed selected.");
		}
		
		if ( displayDataBusMode == DisplayDataBusMode.Mode4bits && (pinData4 == null || pinData5 == null || pinData6 == null || pinData7 == null) ){
			throw new DisplayException("databus mode 4bits: not all 4 pins needed selected.");
		}
		
		// Configuration initiale
		sendPulseEnable();
		sendPulseEnable();
		sendPulseEnable();
		sendPulseEnable();
		if ( this.lines <= 2 ){
			functionSet(displayDataBusMode, false, false);
		}
		else{
			functionSet(displayDataBusMode, true, false);
		}
		displayControl(true, false, false);
		entryModeSet(true, false);
		clearDisplay();
	}
	
	private void pauseInstructionCycle() throws DisplayException{
		try {
			Thread.sleep(0, 100000);
		}
		catch (InterruptedException e) {
			throw new DisplayException("error during waiting instruction cycle", e);
		}
	}
	
	private void sendPulseEnable() throws DisplayException{
		try{
			gpioTools.setValue(pinEnable, true);
			pauseInstructionCycle();
			gpioTools.setValue(pinEnable, false);
			pauseInstructionCycle();
		}
		catch(GPIOException e){
			throw new DisplayException("error using GPIO during sending pulse to enable", e);
		}
	}
	
	private void clearDataBus() throws DisplayException{
		try{
			if ( displayDataBusMode == DisplayDataBusMode.Mode8bits ){
				gpioTools.setValue(pinData0, false);
				gpioTools.setValue(pinData1, false);
				gpioTools.setValue(pinData2, false);
				gpioTools.setValue(pinData3, false);
			}
			gpioTools.setValue(pinData4, false);
			gpioTools.setValue(pinData5, false);
			gpioTools.setValue(pinData6, false);
			gpioTools.setValue(pinData7, false);
		}
		catch(GPIOException e){
			throw new DisplayException("error using GPIO during sending data to display databus", e);
		}
	}
	
	public boolean checkBusyFlag() throws DisplayException{
		boolean result = false;
		try{
			gpioTools.setValue(pinRegisterSelect, false); // Busy flag and adresse counter
			gpioTools.setValue(pinReadWrite, true); // En lecture
			
			if ( displayDataBusMode == DisplayDataBusMode.Mode8bits ){
				gpioTools.setDirection(pinData0, GPIODirection.in);
				gpioTools.setDirection(pinData1, GPIODirection.in);
				gpioTools.setDirection(pinData2, GPIODirection.in);
				gpioTools.setDirection(pinData3, GPIODirection.in);
			}
			gpioTools.setDirection(pinData4, GPIODirection.in);
			gpioTools.setDirection(pinData5, GPIODirection.in);
			gpioTools.setDirection(pinData6, GPIODirection.in);
			gpioTools.setDirection(pinData7, GPIODirection.in);
			
			gpioTools.setValue(pinEnable, true);
			pauseInstructionCycle();
			
			result = gpioTools.getValueBoolean(pinData7);
			
			gpioTools.setValue(pinEnable, false);
			pauseInstructionCycle();
			
			if ( displayDataBusMode == DisplayDataBusMode.Mode4bits ){
				sendPulseEnable();
			}
		}
		catch(GPIOException e){
			throw new DisplayException("error using GPIO during sending data to display databus", e);
		}
		return result;
	}
	
	private void sendByteToDataBus8bits(byte code) throws DisplayException{
		try{
			gpioTools.setDirection(pinData0, GPIODirection.out);
			gpioTools.setDirection(pinData1, GPIODirection.out);
			gpioTools.setDirection(pinData2, GPIODirection.out);
			gpioTools.setDirection(pinData3, GPIODirection.out);
			gpioTools.setDirection(pinData4, GPIODirection.out);
			gpioTools.setDirection(pinData5, GPIODirection.out);
			gpioTools.setDirection(pinData6, GPIODirection.out);
			gpioTools.setDirection(pinData7, GPIODirection.out);
			
			gpioTools.setValue(pinData0, (code & 1) > 0 );
			gpioTools.setValue(pinData1, (code & 2) > 0 );
			gpioTools.setValue(pinData2, (code & 4) > 0 );
			gpioTools.setValue(pinData3, (code & 8) > 0 );
			gpioTools.setValue(pinData4, (code & 16) > 0 );
			gpioTools.setValue(pinData5, (code & 32) > 0 );
			gpioTools.setValue(pinData6, (code & 64) > 0 );
			gpioTools.setValue(pinData7, (code & 128) > 0 );
		}
		catch(GPIOException e){
			throw new DisplayException("error using GPIO during sending data to display databus", e);
		}
	}
	
	private void sendByteToDataBus4bits(byte code, boolean lsb) throws DisplayException{
		try{
			gpioTools.setDirection(pinData4, GPIODirection.out);
			gpioTools.setDirection(pinData5, GPIODirection.out);
			gpioTools.setDirection(pinData6, GPIODirection.out);
			gpioTools.setDirection(pinData7, GPIODirection.out);
			
			if ( lsb ){
				gpioTools.setValue(pinData4, (code & 1) > 0 );
				gpioTools.setValue(pinData5, (code & 2) > 0 );
				gpioTools.setValue(pinData6, (code & 4) > 0 );
				gpioTools.setValue(pinData7, (code & 8) > 0 );
			}
			else{
				gpioTools.setValue(pinData4, (code & 16) > 0 );
				gpioTools.setValue(pinData5, (code & 32) > 0 );
				gpioTools.setValue(pinData6, (code & 64) > 0 );
				gpioTools.setValue(pinData7, (code & 128) > 0 );
			}
		}
		catch(GPIOException e){
			throw new DisplayException("error using GPIO during sending data to display databus", e);
		}
	}
	
	public void sendInstruction(byte code) throws DisplayException{
		try{
			gpioTools.setValue(pinRegisterSelect, false); // Registre instruction
			gpioTools.setValue(pinReadWrite, false); // En écriture
			clearDataBus();
			
			if ( displayDataBusMode == DisplayDataBusMode.Mode8bits ){
				// 8 bits mode
				sendByteToDataBus8bits(code);
				sendPulseEnable();
			}
			else{
				// 4 bits mode
				sendByteToDataBus4bits(code, false);
				sendPulseEnable();
				sendByteToDataBus4bits(code, true);
				sendPulseEnable();
			}
			
			if ( !ignoreIfBusy ){
				int waitingBusyFlag = 0;
				while ( checkBusyFlag() && waitingBusyFlag < 100 ){
					waitingBusyFlag++;
				}
				if ( waitingBusyFlag == 100 ){
					throw new DisplayException("weird ... display seems to be too long to execute insctruction.");
				}
			}
			
			clearDataBus();
		}
		catch(GPIOException e){
			throw new DisplayException("error during sending instruction", e);
		}
	}
	
	public void sendData(byte code) throws DisplayException{
		try{
			gpioTools.setValue(pinRegisterSelect, true); // Registre data
			gpioTools.setValue(pinReadWrite, false); // En écriture
			clearDataBus();
			
			if ( displayDataBusMode == DisplayDataBusMode.Mode8bits ){
				// 8 bits mode
				sendByteToDataBus8bits(code);
				sendPulseEnable();
			}
			else{
				// 4 bits mode
				sendByteToDataBus4bits(code, false);
				sendPulseEnable();
				sendByteToDataBus4bits(code, true);
				sendPulseEnable();
			}
			
			if ( !ignoreIfBusy ){
				int waitingBusyFlag = 0;
				while ( checkBusyFlag() && waitingBusyFlag < 100 ){
					waitingBusyFlag++;
				}
				if ( waitingBusyFlag == 100 ){
					throw new DisplayException("weird ... display seems to be too long to accept data.");
				}
			}
			
			clearDataBus();
		}
		catch(GPIOException e){
			throw new DisplayException("error during sending instruction", e);
		}
	}
	
	public void clearDisplay() throws DisplayException{
		sendInstruction((byte)0x01);
	}
	
	public void displayControl(boolean displayOnOff, boolean cursorPositionOnOff, boolean cursorBlinkingOnOff) throws DisplayException{
		this.displayOnOff = displayOnOff;
		this.cursorPositionOnOff = cursorPositionOnOff;
		this.cursorBlinkingOnOff = cursorBlinkingOnOff;
		
		byte instruction = 0x08;
		if ( displayOnOff ){
			instruction += 0x04;
		}
		if ( cursorPositionOnOff ){
			instruction += 0x02;
		}
		if ( cursorBlinkingOnOff ){
			instruction += 0x01;
		}
		sendInstruction(instruction);
	}
	
	public void displayControlDisplay(boolean displayOnOff) throws DisplayException{
		displayControl(displayOnOff, this.cursorPositionOnOff, this.cursorBlinkingOnOff);
	}
	
	public void displayControlCursorPosition(boolean cursorPositionOnOff) throws DisplayException{
		displayControl(this.displayOnOff, cursorPositionOnOff, this.cursorBlinkingOnOff);
	}
	
	public void displayControlCursorBlinking(boolean cursorBlinkingOnOff) throws DisplayException{
		displayControl(this.displayOnOff, this.cursorPositionOnOff, cursorBlinkingOnOff);
	}
	
	public void functionSet(DisplayDataBusMode displayDataBusMode, boolean numberDisplayedLines, boolean characterFont) throws DisplayException{
		byte instruction = 0x20;
		if ( displayDataBusMode == DisplayDataBusMode.Mode8bits ){
			instruction += 0x10;
		}
		if ( numberDisplayedLines ){
			instruction += 0x08;
		}
		if ( characterFont ){
			instruction += 0x04;
		}
		sendInstruction(instruction);
	}
	
	public void entryModeSet(boolean direction, boolean accompaniesDisplayShift) throws DisplayException{
		byte instruction = 0x04;
		if ( direction ){
			instruction += 0x02;
		}
		if ( accompaniesDisplayShift ){
			instruction += 0x01;
		}
		sendInstruction(instruction);
	}
	
	public void setDDRAMAddress(byte address) throws DisplayException{
		byte instruction = (byte) 0x80;
		instruction += address;
		sendInstruction(instruction);
	}
	
	public void print(String text) throws DisplayException{
		char carac;
		for ( int i = 0 ; i < text.length() ; i++ ){
			carac = text.charAt(i);
			this.sendData((byte) carac);
		}
	}
	
	public void locate(int line, int column) throws DisplayException{
		if ( line < 0 || line >= lines ){
			throw new DisplayException("wrong line number");
		}
		if ( column < 0 || column >= cols ){
			throw new DisplayException("wrong column number");
		}
		byte ramPosition = 0;
		if ( line == 1 ){
			ramPosition = 0x40;
		}
		if ( line == 2 ){
			ramPosition = 0x14;
		}
		if ( line == 3 ){
			ramPosition = 0x54;
		}
		ramPosition += column;
		setDDRAMAddress(ramPosition);
	}
}
