package fr.jelora.jraspitools.gpio;

import java.io.File;
import java.io.FileInputStream;
import java.io.PrintWriter;

import org.apache.commons.io.IOUtils;

import fr.jelora.jraspitools.RasPiPin;

public class GPIOTools {
	
	public boolean init(RasPiPin pin, GPIODirection direction) throws GPIOException{
		return this.init(pin, direction, false, false);
	}
	
	public boolean init(RasPiPin pin, GPIODirection direction, boolean value) throws GPIOException{
		return this.init(pin, direction, false, value);
	}
	
	public boolean init(RasPiPin pin, GPIODirection direction, boolean activeLow, boolean value) throws GPIOException{
		boolean gpioInitialise = false;
		if ( pin != null && direction != null ){
			if ( pin.getIdGPIO() > 0 ){
				File fileExport = new File("/sys/class/gpio/export");
				if ( fileExport.exists() ){
					if ( fileExport.canWrite() ){
						try{
							PrintWriter writerExport = new PrintWriter(fileExport);
							writerExport.print(pin.getIdGPIO() + "");
							writerExport.close();
						}
						catch(Exception e){
							throw new GPIOException("can't open GPIO. Need root ?", e);
						}
					}
					else{
						throw new GPIOException("can't create GPIO. Need root ?");
					}
					
					File dossierGPIO = new File("/sys/class/gpio/gpio" + pin.getIdGPIO());
					if ( dossierGPIO.exists() ){
						if ( dossierGPIO.isDirectory() ){
							this.setDirection(pin, direction);
							this.setActiveLow(pin, activeLow);
							this.setValue(pin, value);
							gpioInitialise = true;
						}
					}
					else{
						throw new GPIOException("GPIO not found");
					}
				}
				else{
					throw new GPIOException("can't open GPIO. /sys/class/gpio/export not found.");
				}
			}
			else{
				throw new GPIOException("not a GPIO");
			}
		}
		else{
			throw new GPIOException("pin/direction null");
		}
		return gpioInitialise;
	}
	
	public void setDirection(RasPiPin pin, GPIODirection direction) throws GPIOException{
		if ( pin != null && direction != null ){
			if ( pin.getIdGPIO() > 0 ){
				File fileDirection = new File("/sys/class/gpio/gpio" + pin.getIdGPIO() + "/direction");
				if ( fileDirection.exists() ){
					try{
						PrintWriter writerDirection = new PrintWriter(fileDirection);
						writerDirection.print(direction.toString());
						writerDirection.close();
					}
					catch(Exception e){
						throw new GPIOException("can't set direction GPIO. Need root ?", e);
					}
				}
				else{
					throw new GPIOException("GPIO not found");
				}
			}
			else{
				throw new GPIOException("not a GPIO");
			}
		}
		else{
			throw new GPIOException("pin/direction null");
		}
	}
	
	public void setActiveLow(RasPiPin pin, boolean activeLow) throws GPIOException{
		if ( pin != null ){
			if ( pin.getIdGPIO() > 0 ){
				File fileActiveLow = new File("/sys/class/gpio/gpio" + pin.getIdGPIO() + "/active_low");
				if ( fileActiveLow.exists() ){
					try{
						PrintWriter writerActiveLow = new PrintWriter(fileActiveLow);
						if ( activeLow ){
							writerActiveLow.print("1");
						}
						else{
							writerActiveLow.print("0");
						}
						writerActiveLow.close();
					}
					catch(Exception e){
						throw new GPIOException("can't set active low GPIO. Need root ?", e);
					}
				}
				else{
					throw new GPIOException("GPIO not found");
				}
			}
			else{
				throw new GPIOException("not a GPIO");
			}
		}
		else{
			throw new GPIOException("pin null");
		}
	}
	
	public void setValue(RasPiPin pin, boolean value) throws GPIOException{
		if ( pin != null ){
			if ( pin.getIdGPIO() > 0 ){
				File fileValue = new File("/sys/class/gpio/gpio" + pin.getIdGPIO() + "/value");
				if ( fileValue.exists() ){
					try{
						PrintWriter writerActiveLow = new PrintWriter(fileValue);
						if ( value ){
							writerActiveLow.print("1");
						}
						else{
							writerActiveLow.print("0");
						}
						writerActiveLow.close();
					}
					catch(Exception e){
						throw new GPIOException("can't set value GPIO. Need root ?", e);
					}
				}
				else{
					throw new GPIOException("GPIO not found");
				}
			}
			else{
				throw new GPIOException("not a GPIO");
			}
		}
		else{
			throw new GPIOException("pin null");
		}
	}
	
	public boolean getValueBoolean(RasPiPin pin) throws GPIOException{
		boolean value = false;
		Integer valueInteger = this.getValueInteger(pin);
		if ( valueInteger != null ){
			if ( valueInteger == 1 ){
				value = true;
			}
		}
		return value;
	}
	
	public int getValueInteger(RasPiPin pin) throws GPIOException{
		int value;
		if ( pin != null ){
			if ( pin.getIdGPIO() > 0 ){
				File fileValue = new File("/sys/class/gpio/gpio" + pin.getIdGPIO() + "/value");
				if ( fileValue.exists() ){
					try{
						FileInputStream inputStream = new FileInputStream(fileValue);
						value = Integer.parseInt(IOUtils.toString(inputStream).trim());
					}
					catch(Exception e){
						throw new GPIOException("can't read GPIO. Need root ?", e);
					}
				}
				else{
					throw new GPIOException("GPIO not found");
				}
			}
			else{
				throw new GPIOException("not a GPIO");
			}
		}
		else{
			throw new GPIOException("pin null");
		}
		return value;
	}
	
	public void closeGPIO(RasPiPin pin) throws GPIOException{
		if ( pin != null ){
			if ( pin.getIdGPIO() > 0 ){
				File fileUnexport = new File("/sys/class/gpio/unexport");
				if ( fileUnexport.exists() ){
					if ( fileUnexport.canWrite() ){
						try{
							PrintWriter writerUnexport = new PrintWriter(fileUnexport);
							writerUnexport.print(pin.getIdGPIO() + "");
							writerUnexport.close();
						}
						catch(Exception e){
							throw new GPIOException("can't close GPIO. Need root ?", e);
						}
					}
					else{
						throw new GPIOException("can't close GPIO. Need root ?");
					}
				}
			}
			else{
				throw new GPIOException("not a GPIO");
			}
		}
		else{
			throw new GPIOException("pin null");
		}
	}
}
